/// This is the simplest possible client using rustls that does something useful:
/// it accepts the default configuration, loads some root certs, and then connects
/// to google.com and issues a basic HTTP request.  The response is printed to stdout.
///
/// It makes use of rustls::Stream to treat the underlying TLS connection as a basic
/// bi-directional stream -- the underlying IO is performed transparently.
///
/// Note that `unwrap()` is used to deal with networking errors; this is not something
/// that is sensible outside of example code.
use std::sync::Arc;

use log::{error, info, trace};
use managesieve::{Capability, MSResult, Response};
use nom::{error::ParseError, IResult};
use sieverman::parser::parse_server_config;
use std::fmt::Debug;
use std::net::ToSocketAddrs;
use tokio::{
    io::{self, split, AsyncBufReadExt, AsyncWriteExt, BufReader},
    net::TcpStream,
};
use tokio_rustls::{
    rustls::{ClientConfig, KeyLogFile, RootCertStore},
    TlsConnector,
};

#[macro_export]
macro_rules! parse_response {
    ($reader:expr, $parser:expr, $dst:expr) => {{
        let mut buf = String::new();
        let capabilities = 'read_more: loop {
            trace!("reading");
            let read = $reader.read_line(&mut buf).await;
            if let Ok(num) = read {
                trace!("Read {} bytes", num);
                trace!("Read Buffer: {:?}", &read);
                trace!("Read Buffer: {:?}", &buf);
                $dst += &buf;
                buf.clear();
            };
            match $parser(&$dst) {
                read @ Ok((_, _)) => {
                    //trace!("{:?}", &response);
                    break 'read_more read;
                }
                Err(_e) => continue,
                Err(e) => {
                    error!("invalid response");
                    break Err(e);
                }
                Err(e) => {
                    error!("invalid input");
                    break Err(e);
                }
            };
        };
        capabilities
    }};
}

async fn run() -> io::Result<()> {
    let mut root_store = RootCertStore::empty();
    root_store.extend(webpki_roots::TLS_SERVER_ROOTS.iter().cloned());
    let mut config = ClientConfig::builder()
        .with_root_certificates(root_store)
        .with_no_client_auth();

    // Allow using SSLKEYLOGFILE.
    config.key_log = Arc::new(KeyLogFile::new());
    let connector = TlsConnector::from(Arc::new(config));
    let server_address = ("teilgedanken.de", 4190).to_socket_addrs()?.next().unwrap();
    let stream = TcpStream::connect(&server_address).await?;
    let domain = pki_types::ServerName::try_from("teilgedanken.de").unwrap();
    let stream = connector.connect(domain, stream).await?;
    let (reader, mut writer) = split(stream);
    let mut reader = BufReader::new(reader);
    let mut dst = String::new();
    let mut buf = String::new();
    let capabilities: Option<(&str, Vec<Capability>)> = 'read_more: loop {
        trace!("reading");
        let read = reader.read_line(&mut buf).await;
        if let Ok(num) = read {
            trace!("Read {} bytes", num);
            trace!("Read Buffer: {:?}", &read);
            trace!("Read Buffer: {:?}", &buf);
            dst += &buf;
            buf.clear();
        };
        match managesieve::response_capability(&dst) {
            Ok((rest, capability, response)) => {
                trace!("{:?}", response);
                break 'read_more Some((rest, capability));
            }
            Err(managesieve::MSError::IncompleteResponse) => continue,
            Err(managesieve::MSError::InvalidResponse) => {
                error!("invalid response");
                break None;
            }
            Err(managesieve::MSError::InvalidInput) => {
                error!("invalid input");
                break None;
            }
        };
    };
    if let Some((rest, capability)) = capabilities {
        trace!("Parsed: \n{}", dst);
        dst = rest.to_string();
        info!("Rest content:\n{:?}", &dst);
        info!("The server config: {:?}", capability);
        let have_space_command = managesieve::Command::have_space("loser.test", 249)
            .expect("failed to create have_space command");
        writer
            .write_all(have_space_command.to_string().as_bytes())
            .await?;
        let response_parser = managesieve::response_havespace;
        let result = parse_response!(reader, response_parser, dst);
        dbg!(result);
        let auth_command = managesieve::Command::authenticate();
        writer
            .write_all(auth_command.to_string().as_bytes())
            .await?;
        let response_parser = managesieve::response_authenticate;
        let result = parse_response!(reader, response_parser, dst);
        dbg!(result);
    }
    Ok(())
}

async fn read_until_complete<'a, Reader, Target>(
    mut reader: Reader,
    mut parser: impl FnMut(&str) -> IResult<&'a str, (Target, Response)>,
) -> IResult<&'a str, Target>
where
    Reader: AsyncBufReadExt + Unpin,
    Target: Debug,
{
    let mut dst = String::new();
    let mut buf = String::new();
    let capabilities: MSResult<Target> = 'read_more: loop {
        trace!("reading");
        let read = reader.read_line(&mut buf).await;
        if let Ok(num) = read {
            trace!("Read {} bytes", num);
            trace!("Read Buffer: {:?}", &read);
            trace!("Read Buffer: {:?}", &buf);
            dst += &buf;
            buf.clear();
        };
        match parser(&dst) {
            read @ Ok((_, _)) => {
                //trace!("{:?}", &response);
                break 'read_more read;
            }
            Err(_e) => continue,
            Err(e) => {
                error!("invalid response");
                break Err(e);
            }
            Err(e) => {
                error!("invalid input");
                break Err(e);
            }
        };
    };
    if let Ok((rest, (target, response))) = capabilities {
        trace!("Parsed: \n{}", dst);
        dst = rest.to_string();
        trace!("Rest content:\n{:?}", &dst);
        trace!("Response content: {:?}", &response);
        info!("The server config: {:?}", target);
    }
    todo!()
}

fn main() {
    env_logger::init();
    tokio::runtime::Builder::new_current_thread()
        .enable_io()
        .build()
        .unwrap()
        .block_on(run())
        .expect("Failed to run or interrupted");
}
