use std::{any, net::SocketAddr, sync::Arc};

use log::{error, info, trace};
use managesieve::Capability;
use std::net::ToSocketAddrs;
use thiserror::Error;
use tokio::{
    io::{split, AsyncBufReadExt, BufReader, ReadHalf, WriteHalf},
    net::TcpStream,
    sync::Mutex,
    task::JoinHandle,
};
use tokio_rustls::{
    client::TlsStream,
    rustls::{ClientConfig, KeyLogFile, RootCertStore},
    TlsConnector,
};

pub mod parser;

#[derive(Debug, Clone)]
pub struct ConnectionInfo {
    domain: String,
    port: u16,
}

impl ConnectionInfo {
    pub fn new(domain: &str, port: u16) -> Self {
        Self {
            domain: domain.to_string(),
            port,
        }
    }
    pub async fn connect(self) -> Result<ConnectionConnected, anyhow::Error> {
        trace!("Building Cert Store");
        let mut root_store = RootCertStore::empty();
        root_store.extend(webpki_roots::TLS_SERVER_ROOTS.iter().cloned());
        let mut config = ClientConfig::builder()
            .with_root_certificates(root_store)
            .with_no_client_auth();

        // Allow using SSLKEYLOGFILE.
        config.key_log = Arc::new(KeyLogFile::new());
        trace!("Creating connector");
        let connector = TlsConnector::from(Arc::new(config));
        trace!("Connecting to the server.");
        let stream = TcpStream::connect((self.domain.clone(), self.port)).await?;
        let domain = pki_types::ServerName::try_from(self.domain.clone())?.to_owned();
        trace!("Attach the connector");
        let stream = connector.connect(domain, stream).await?;
        let (reader, mut writer) = split(stream);
        let buffer = Arc::new(Mutex::new(String::new()));
        let spawned_buffer = buffer.clone();

        let join_handle = tokio::spawn(async move {
            let mut buf = String::new();
            let mut reader = BufReader::new(reader);
            loop {
                trace!("Waiting for more data to read…");
                let read = reader.read_line(&mut buf).await;
                if let Ok(num) = read {
                    trace!("Read {} bytes", num);
                    trace!("Read Buffer: {:?}", &buf);
                    {
                        let mut s = spawned_buffer.lock().await;
                        s.push_str(&buf);
                    }
                    buf.clear();
                };
            }
        });
        Ok(ConnectionConnected {
            info: self,
            writer: Arc::new(writer), //server_settings: todo!(),
            buffer,
            server_settings: Vec::new(), //join_handle: Arc::new(join_handle),
        })
    }
}

#[derive(Debug, Clone)]
pub struct ConnectionConnected {
    info: ConnectionInfo,
    //reader: Arc<ReadHalf<TlsStream<TcpStream>>>,
    writer: Arc<WriteHalf<TlsStream<TcpStream>>>,
    buffer: Arc<Mutex<String>>,
    //pub join_handle: Arc<JoinHandle<()>>,
    server_settings: Vec<Capability>,
}

#[derive(Error, Debug)]
pub enum ReadError {
    #[error("The data entered was invalid")]
    InvalidInput(#[source] managesieve::Error),
    #[error("The data the server responded was invalid")]
    InvalidResponse(#[source] managesieve::Error),
}

pub enum IsComplete<T> {
    Yes(T),
    No(T),
}

impl ConnectionConnected {
    pub async fn log_buffer(&self) {
        trace!("Buffer is: \n{}", self.buffer.lock().await)
    }
    pub async fn log_server_settings(&self) {
        trace!("Serversettings:\n{:?}", self.server_settings)
    }
    pub async fn read_introduction(self) -> Result<IsComplete<Self>, anyhow::Error> {
        let Self {
            info,
            //reader: Arc<ReadHalf<TlsStream<TcpStream>>>,
            writer,
            buffer,
            server_settings,
        } = self;
        let mut bf = buffer.lock().await;
        let bf_str = bf.clone();
        match managesieve::response_capability(&bf_str) {
            Ok((rest, capability, response)) => {
                trace!("{:?}", response);
                bf.clear();
                bf.push_str(rest);
                drop(bf);
                Ok(IsComplete::Yes(Self {
                    info,
                    writer: writer.clone(),
                    buffer: buffer.clone(),
                    server_settings: capability,
                }))
            }
            Err(managesieve::Error::IncompleteResponse) => Ok(IsComplete::No(Self {
                info,
                writer,
                buffer: buffer.clone(),
                server_settings,
            })),
            Err(managesieve::Error::InvalidResponse) => {
                error!("invalid response");
                Err(managesieve::Error::InvalidResponse)?
            }
            Err(managesieve::Error::InvalidInput) => {
                error!("invalid input");
                Err(managesieve::Error::InvalidResponse)?
            }
        }
    }
    pub fn get_greeting(&self) -> Option<String> {
        for c in self.server_settings.iter() {
            if let Capability::Implementation(s) = c {
                return Some(s.clone());
            }
        }
        None
    }
}

#[derive(Debug, Clone)]
pub struct ConnectionAuthenticated {
    info: ConnectionInfo,
    server_settings: ServerSettings,
}

#[derive(Debug, Clone)]
pub struct ServerSettings {
    pub implementation: String,
    pub version: Vec<i32>,
    pub sasl: Vec<Methods>,
    pub capabilities: Vec<String>,
    pub notify: Vec<String>,
    pub max_redirects: u32,
}

#[derive(Debug, Clone, Copy)]
pub enum Methods {
    Plain,
    OAuthbearer,
}
