use std::sync::Arc;

use tokio::io::AsyncWriteExt;
use tokio::net::TcpStream;

async fn connect(addr: &str) -> Result<TlsStream<TcpStream>, Box<dyn std::error::Error>> {
    let mut root_cert_store = RootCertStore::empty();

    for cert in rustls_native_certs::load_native_certs().expect("could not load platform certs") {
        root_cert_store.add(cert)
    }
    let roots = webpki_roots::TLS_SERVER_ROOTS.iter().map(|ta| {
        OwnedTrustAnchor::from_subject_spki_name_constraints(
            ta.subject,
            ta.subject_public_key_info,
            ta.name_constraints,
        )
    });
    root_cert_store.add_trust_anchors(roots);
    let config = ClientConfig::builder()
        .with_safe_defaults()
        .with_root_certificates(root_cert_store)
        .with_no_client_auth();
    let connector = TlsConnector::from(Arc::new(config));
    let dnsname = ServerName::try_from("www.rust-lang.org").unwrap();

    let stream = TcpStream::connect(&addr).await?;
    let mut stream = connector.connect(dnsname, stream).await?;

    Ok(tls_stream)
}

async fn run() -> Result<(), Box<dyn std::error::Error>> {
    let tls_stream = connect("127.0.0.1:8080").await?;

    // Send a request to the server
    let request = b"GET / HTTP/1.1\r\nHost: example.com\r\n\r\n";
    tls_stream.write_all(request)?;

    // Read the response from the server
    let mut buf = [0; 1024];
    let n = tls_stream.read(&mut buf)?;
    println!("{}", String::from_utf8_lossy(&buf[..n]));

    Ok(())
}

fn main() {
    tokio::runtime::Builder::new()
        .threaded_scheduler()
        .build()
        .unwrap()
        .block_on(run());
}
